using System;
using System.Collections.Generic;

namespace id.Quake.Files {
	public partial class QuakeC {


		private int StatementPointer;
		private Function WorkingFunction;

		private struct StackEntry {
			public int StatementPointer;
			public Function Function;
			public StackEntry(int statementPointer, Function function) {
				this.StatementPointer = statementPointer;
				this.Function = function;

			}
		}

		private Stack<StackEntry> FunctionStack;
		private Stack<int> LocalStack;

		private void EnterFunction(Function function) {

			// Preserve locals:
			for (uint i = 0, p = function.FirstParameter; i < function.Locals; ++i) {
				LocalStack.Push(this.GetGlobal(p * 4).Integer);
				++p;
			}

			// Copy parameters:
			for (uint i = 0, p = function.FirstParameter; i < function.ParameterCount; ++i) {
				for (uint j = 0; j < function.ParameterSizes[i]; ++j) {
					this.GetGlobal(p * 4).Integer = this.GetGlobal(4 * ((uint)GlobalVariableOffsets.Parameter0 + i * 3 + j)).Integer;
					++p;
				}
			}

			this.FunctionStack.Push(new StackEntry(this.StatementPointer, this.WorkingFunction));
			this.StatementPointer = function.FirstStatement;
			this.WorkingFunction = function;
		}

		private void LeaveFunction() {

			// Restore locals:
			for (uint i = 0, p = this.WorkingFunction.FirstParameter + this.WorkingFunction.Locals; i < this.WorkingFunction.Locals; ++i) {
				--p;
				this.GetGlobal(p * 4).Integer = LocalStack.Pop();
			}

			StackEntry SE = this.FunctionStack.Pop();
			this.WorkingFunction = SE.Function;
			this.StatementPointer = SE.StatementPointer;
		}

		public void Invoke(EdictCollection.Edict edict, Function method) {
			this.Self = edict;
			this.Execute(method);
		}


	}
}
