using System;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class QuakeC {

		public class FunctionCollection {

			private readonly Function[] Functions;

			/// <summary>Gets a function based on its name.</summary>
			/// <param name="name">The name of the function to retrieve.</param>
			public Function this[string name] {
				get {
					foreach (Function Func in this.Functions) {
						if (Func.Name == name) return Func;						
					}
					throw new InvalidOperationException("Function " + name + " not defined.");
				}
			}

			/// <summary>Gets a function based on its index.</summary>
			/// <param name="name">The index of the function to retrieve.</param>
			public Function this[int index] {
				get {
					return this.Functions[index];
				}
			}

			/// <summary>Gets the number of functions.</summary>
			public int Count {
				get {
					return this.Functions.Length;
				}
			}

			public FunctionCollection(Function[] functions) {
				this.Functions = functions;
			}

		}

		/// <summary>Describes a function.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Function {

			private readonly QuakeC Source;

			public readonly int FirstStatement;
			public readonly int FirstParameter;
			public readonly int Locals;
			public readonly int Profile;
			public readonly int ParameterCount;
			public readonly byte[] ParameterSizes;
			public readonly int Index;

			#region Properties

			private readonly int name;
			/// <summary>Gets the name of the function.</summary>
			public string Name { get { return this.Source.GetString(this.name); } }

			private readonly int sourceFile;
			/// <summary>Gets the name of the function.</summary>
			public string SourceFile { get { return this.Source.GetString(this.sourceFile); } }


			public override string ToString() {
				if (string.IsNullOrEmpty(this.Name)) {
					return "null::void()";
				} else {
					return this.SourceFile + "::" + this.Name + "()";
				}
			}

			#endregion


			public Function(QuakeC source, Stream stream, int index) {
				this.Source = source;
				this.Index = index;
				BinaryReader FunctionReader = new BinaryReader(stream);
				this.FirstStatement = FunctionReader.ReadInt32();
				this.FirstParameter = FunctionReader.ReadInt32();
				this.Locals = FunctionReader.ReadInt32();
				this.Profile = FunctionReader.ReadInt32();
				this.name = FunctionReader.ReadInt32();
				this.sourceFile = FunctionReader.ReadInt32();
				this.ParameterCount = FunctionReader.ReadInt32();
				this.ParameterSizes = FunctionReader.ReadBytes(8);

			}

		}

	}
}
