using System;
using QuakeC = id.Quake.Files.QuakeC;

namespace id.Quake {
	public partial class VirtualMachine {

		private Random RandomNumberSource = new Random();

		#region Delegate Table

		delegate void BuiltInFunction();
		BuiltInFunction[] BuiltInFunctions;

		private void InitialiseBuiltInFunctions() {

			this.BuiltInFunctions = new BuiltInFunction[]{
				PF_Fixme,
				PF_makevectors,	// void(entity e)	makevectors 		= #1;
				PF_setorigin,	// void(entity e, vector o) setorigin	= #2;
				PF_setmodel,	// void(entity e, string m) setmodel	= #3;
				PF_setsize,	// void(entity e, vector min, vector max) setsize = #4;
				PF_Fixme,	// void(entity e, vector min, vector max) setabssize = #5;
				PF_break,	// void() break						= #6;
				PF_random,	// float() random						= #7;
				PF_sound,	// void(entity e, float chan, string samp) sound = #8;
				PF_normalize,	// vector(vector v) normalize			= #9;
				PF_error,	// void(string e) error				= #10;
				PF_objerror,	// void(string e) objerror				= #11;
				PF_vlen,	// float(vector v) vlen				= #12;
				PF_vectoyaw,	// float(vector v) vectoyaw		= #13;
				PF_Spawn,	// entity() spawn						= #14;
				PF_Remove,	// void(entity e) remove				= #15;
				PF_traceline,	// float(vector v1, vector v2, float tryents) traceline = #16;
				PF_checkclient,	// entity() clientlist					= #17;
				PF_Find,	// entity(entity a, .string fld, string match) find = #18;
				PF_precache_sound,	// void(string s) precache_sound		= #19;
				PF_precache_model,	// void(string s) precache_model		= #20;
				PF_stuffcmd,	// void(entity client, string s)stuffcmd = #21;
				PF_findradius,	// entity(vector org, float rad) findradius = #22;
				PF_bprint,	// void(string s) bprint				= #23;
				PF_sprint,	// void(entity client, string s) sprint = #24;
				PF_dprint,	// void(string s) dprint				= #25;
				PF_ftos,	// void(string s) ftos				= #26;
				PF_vtos,	// void(string s) vtos				= #27;
				PF_coredump,
				PF_traceon,
				PF_traceoff,
				PF_eprint,	// void(entity e) debug print an entire entity
				PF_walkmove, // float(float yaw, float dist) walkmove
				PF_Fixme, // float(float yaw, float dist) walkmove
				PF_droptofloor,
				PF_lightstyle,
				PF_rint,
				PF_floor,
				PF_ceil,
				PF_Fixme,
				PF_checkbottom,
				PF_pointcontents,
				PF_Fixme,
				PF_fabs,
				PF_aim,
				PF_cvar,
				PF_localcmd,
				PF_nextent,
				PF_particle,
				PF_changeyaw,
				PF_Fixme,
				PF_vectoangles,

				PF_WriteByte,
				PF_WriteChar,
				PF_WriteShort,
				PF_WriteLong,
				PF_WriteCoord,
				PF_WriteAngle,
				PF_WriteString,
				PF_WriteEntity,

				PF_sin,
				PF_cos,
				PF_sqrt,
				PF_changepitch,
				PF_TraceToss,
				PF_etos,
				PF_WaterMove,

				SV_MoveToGoal,
				PF_precache_file,
				PF_makestatic,

				PF_changelevel,
				PF_Fixme,

				PF_cvar_set,
				PF_centerprint,

				PF_ambientsound,

				PF_precache_model,
				PF_precache_sound,
				PF_precache_file,

				PF_setspawnparms
			};
		}


		#endregion

		void PF_Fixme() { throw new InvalidOperationException(); }

		void PF_makevectors() {
			Vector3 Forward, Right, Up;
			Vector3.GetAngleVectors(Parameter0.Vector, out Forward, out Right, out Up);
			this.quakeCSource.Globals["v_forward"].Value = Forward;
			this.quakeCSource.Globals["v_right"].Value = Right;
			this.quakeCSource.Globals["v_up"].Value = Up;
		}

		void PF_setorigin() {
			//TODO: Reposition entity.
			Parameter0.Edict.Properties["origin"] = Parameter1.Vector;
		}

		private void UpdateSize(Edict ed, BoundingBox box) {
			ed.Properties["mins"] = box.Min;
			ed.Properties["maxs"] = box.Max;
			ed.Properties["size"] = box.Max - box.Min;
		}

		/// <summary>setmodel(entity, model)</summary>
		void PF_setmodel() {

			Edict Ed = this.Parameter0.Edict;
			string ModelName = this.Parameter1.String;

			//TODO: Check this!
			if (string.IsNullOrEmpty(ModelName)) return;

			int ModelIndex;
			if (!this.ModelCacheIndex.TryGetValue(ModelName, out ModelIndex)) {
				throw new InvalidOperationException("Model '" + ModelName + "' not precached.");
			}

			//HACK: Is this correct?
			if (string.IsNullOrEmpty(ModelName)) return;

			// String copy:
			Ed.Properties["model"] = this.Parameter1.Integer;
			Ed.Properties["modelindex"] = ModelIndex;

			//TODO: Bounding box.
			//BOTCH: "do this right"
			// mod->mins[0] = mod->mins[1] = mod->mins[2] = -16;
			// mod->maxs[0] = mod->maxs[1] = mod->maxs[2] = 16;


			BoundingBox Box = ((Files.IBoundingBoxed)this.ModelCache[ModelIndex]).Bounds;
			Box.Min -= new Vector3(16f);
			Box.Max += new Vector3(16f);
			UpdateSize(Ed, Box);
		}

		/// <summary>
		/// The size box is rotated by the current angle
		/// setsize (entity, minvector, maxvector)
		/// </summary>
		void PF_setsize() {
			Edict Ed = this.Parameter0.Edict;
			BoundingBox Box = new BoundingBox(this.Parameter1.Vector, this.Parameter2.Vector);
			UpdateSize(Ed, Box);
		}

		void PF_break() { throw new InvalidOperationException(); }

		/// <summary>
		/// Returns a number from 0 &lt;= num &lt; 1.
		/// random()
		/// </summary>
		void PF_random() { this.Return.Float = (float)RandomNumberSource.NextDouble(); }

		/// <summary>
		/// PF_sound
		/// Each entity can have eight independant sound sources, like voice, weapon, feet, etc.
		/// Channel 0 is an auto-allocate channel, the others override anything already running on that entity/channel pair.
		/// An attenuation of 0 will play full volume everywhere in the level.
		/// Larger attenuations will drop off.
		/// </summary>
		void PF_sound() {
			//TODO: Play sound.
		}

		void PF_normalize() { throw new InvalidOperationException(); }
		void PF_error() { throw new InvalidOperationException(); }

		/// <summary>
		/// PF_objerror
		/// Dumps out self, then an error message.  The program is aborted and self is removed, but the level can continue.
		/// objerror(value)
		/// </summary>
		void PF_objerror() {
			//TODO: Object error?
		}

		/// <summary>
		/// PF_vlen
		/// scalar vlen(vector)
		/// </summary>
		void PF_vlen() {
			this.Return.Float = this.Parameter0.Vector.Magnitude;
		}


		void PF_vectoyaw() { throw new InvalidOperationException(); }

		void PF_Spawn() {
			Edict Ed = this.Edicts.Allocate();
			this.Return.Edict = Ed;
		}


		void PF_Remove() {
			this.Edicts.Free(this.Parameter0.Edict);
		}

		void PF_traceline() { throw new InvalidOperationException(); }
		void PF_checkclient() { throw new InvalidOperationException(); }

		void PF_Find() {
			Edict StartSearch = this.Parameter0.Edict;
			int FieldOffset  = this.Parameter1.Integer;
			string SearchString = this.Parameter2.String;
			for (int i = StartSearch.Index; i < edicts.Capacity; ++i) {
				if (this.edicts.IsActive(i)) {
					if (this.quakeCSource.GetString(this.Edicts.EdictData[i * this.quakeCSource.EdictSize + FieldOffset].Integer) == SearchString) {
						this.Return.Integer = i;
						return;
					}
				}
			}
			this.Return.Integer = 0;			
		}

		void PF_precache_sound() {
			this.Return.Integer = this.Parameter0.Integer;
			string ModelName = this.Parameter0.String;
			this.OnPrecacheSoundRequested(new PrecacheFileRequestedEventArgs(-1, ModelName));
		}

		void PF_precache_model() {
			this.Return.Integer = this.Parameter0.Integer;
			string ModelName = this.Parameter0.String;
			this.OnPrecacheModelRequested(new PrecacheFileRequestedEventArgs(this.CacheModel(ModelName), ModelName));
		}

		void PF_stuffcmd() { throw new InvalidOperationException(); }
		void PF_findradius() { throw new InvalidOperationException(); }
		void PF_bprint() { throw new InvalidOperationException(); }
		void PF_sprint() { throw new InvalidOperationException(); }


		void PF_dprint() { this.OnDebugStringPrinted(new DebugStringPrintedEventArgs(this.Parameter0.String)); }

		void PF_ftos() {
			float Value = this.Parameter0.Float;
			if (Value == (int)Value) {
				this.quakeCSource.StringBuffer = ((int)Value).ToString();
			} else {
				this.quakeCSource.StringBuffer = Value.ToString();
			}
			this.Return.Integer = -1;
		}
		void PF_vtos() {
			Vector3 V = this.Parameter0.Vector;
			this.quakeCSource.StringBuffer = string.Format("'{0:N1} {1:N1} {2:N1}'", V.X, V.Y, V.Z);
			this.Return.Integer = -1;
		}
		void PF_coredump() { throw new InvalidOperationException(); }
		void PF_traceon() { throw new InvalidOperationException(); }
		void PF_traceoff() { throw new InvalidOperationException(); }
		void PF_eprint() { throw new InvalidOperationException(); }

		void PF_walkmove() {
			//TODO: Move an entity.
		}

		void PF_droptofloor() {
			//TODO: Drop entity to floor.
			Edict Ed = this.SelfEdict;
			Vector3 Source = (Vector3)Ed.Properties["origin"];
			Source += new Vector3(0f, 0f, 0.1f);

			Vector3 Destination = Source - new Vector3(0f, 0f, 256f);
			
			Vector3 AttemptedDestination = Destination;

			AttemptedDestination = this.GetCollision(Source, AttemptedDestination).Position;

			Ed.Properties["origin"] = AttemptedDestination;

			Ed.Properties["flags"] = (float)((Edict.EdictFlags)(float)Ed.Properties["flags"] | Edict.EdictFlags.OnGround);

			this.Return.Boolean = true;
		}

		void PF_lightstyle() { this.OnLightStyleChanged(new LightStyleChangedEventArgs((int)this.Parameter0.Float, this.Parameter1.String)); }
		void PF_rint() { throw new InvalidOperationException(); }
		void PF_floor() { throw new InvalidOperationException(); }
		void PF_ceil() { throw new InvalidOperationException(); }
		void PF_checkbottom() { throw new InvalidOperationException(); }
		void PF_pointcontents() { throw new InvalidOperationException(); }

		void PF_fabs() {
			Return.Float = Math.Abs(Parameter0.Float);
		}

		void PF_aim() { throw new InvalidOperationException(); }
		
		void PF_cvar() {
			this.Return.Float = this.ConsoleVariables[this.Parameter0.String].Float;
		}

		void PF_localcmd() { throw new InvalidOperationException(); }
		void PF_nextent() { throw new InvalidOperationException(); }
		void PF_particle() { throw new InvalidOperationException(); }
		void PF_changeyaw() { throw new InvalidOperationException(); }
		void PF_vectoangles() { throw new InvalidOperationException(); }
		void PF_WriteByte() { throw new InvalidOperationException(); }
		void PF_WriteChar() { throw new InvalidOperationException(); }
		void PF_WriteShort() { throw new InvalidOperationException(); }
		void PF_WriteLong() { throw new InvalidOperationException(); }
		void PF_WriteCoord() { throw new InvalidOperationException(); }
		void PF_WriteAngle() { throw new InvalidOperationException(); }
		void PF_WriteString() { throw new InvalidOperationException(); }
		void PF_WriteEntity() { throw new InvalidOperationException(); }
		void PF_sin() { throw new InvalidOperationException(); }
		void PF_cos() { throw new InvalidOperationException(); }
		void PF_sqrt() { throw new InvalidOperationException(); }
		void PF_changepitch() { throw new InvalidOperationException(); }
		void PF_TraceToss() { throw new InvalidOperationException(); }
		void PF_etos() { throw new InvalidOperationException(); }
		void PF_WaterMove() { throw new InvalidOperationException(); }
		void SV_MoveToGoal() { throw new InvalidOperationException(); }

		void PF_precache_file() {
			this.Return.Integer = this.Parameter0.Integer;
			this.OnPrecacheFileRequested(new PrecacheFileRequestedEventArgs(-1, this.Parameter0.String));
		}



		void PF_makestatic() {
			//TODO: Spawn static edicts.
			Edict Ed = this.Parameter0.Edict;
			this.OnSpawnStaticEntity(new SpawnStaticEntityEventArgs(Ed));
			this.edicts.Free(Ed);
		}

		void PF_changelevel() { throw new InvalidOperationException(); }
		
		void PF_cvar_set() {
			this.ConsoleVariables[this.Parameter0.String].String = this.Parameter1.String;
			this.OnCVarChanged(new CVarChangedEventArgs(this.Parameter0.String, this.Parameter1.String)); 
		}
		
		void PF_centerprint() { throw new InvalidOperationException(); }

		void PF_ambientsound() {
			this.OnSpawnAmbientSound(new SpawnAmbientSoundEventArgs(
				this.Parameter0.Vector,
				this.Parameter1.String,
				this.Parameter2.Float,
				this.Parameter3.Float
			));
		}

		void PF_setspawnparms() { throw new InvalidOperationException(); }

	}
}
