using System;

using id.Quake;
using id.Quake.Files;

using System.Diagnostics;
using System.IO;
using System.Collections.Generic;
using System.Reflection;

namespace TestQuakeC {
	class Program {
		static void Main(string[] args) {


			VirtualMachine.ConsoleVariable CV = VirtualMachine.ConsoleVariable.Parse("host_framerate = {\"host_framerate\",\"0\"};    // set for slow motion");
			Console.WriteLine(CV);

			Console.Title = Assembly.GetExecutingAssembly().GetName().Name;

			// Get application path:
			string ApplicationPath = Path.GetDirectoryName(Process.GetCurrentProcess().MainModule.FileName);

			// Set up the resource loader:
			ResourceLoader Resources = new id.Quake.ResourceLoader(@"C:\Program Files\Quake\ID1");
			Resources.AddStandardPackages();

			// Grab a level:
			Level Level = Resources.LoadObject<Level>("maps/e1m1.bsp");

			// Load the QuakeC application:
			QuakeC Application = Resources.LoadObject<QuakeC>("progtest.dat");

			Console.WriteLine("Loaded Application (CRC=0x{0:X4})", Application.Crc);
			Console.WriteLine("{0} functions made up of {1} statements.", Application.Functions.Count, Application.Statements.Length);
			Console.WriteLine("{0} global definitions, {1} field definitions.", Application.Globals.Count, Application.Fields.Count);

			/*Application.DebugStringPrinted += new QuakeC.DebugStringPrintedEventHandler(delegate(object sender, QuakeC.DebugStringPrintedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.White;
				Console.Write(e);
				Console.ForegroundColor = Old;
			});

			Application.PrecacheFileRequested += new QuakeC.PrecacheFileRequestedEventHandler(delegate(object sender, QuakeC.PrecacheFileRequestedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.Blue;
				Console.WriteLine("Precache file '{0}'", e);
				Console.ForegroundColor = Old;
			});

			Application.PrecacheSoundRequested += new QuakeC.PrecacheFileRequestedEventHandler(delegate(object sender, QuakeC.PrecacheFileRequestedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.Red;
				Console.WriteLine("Precache sound '{0}'", e);
				Console.ForegroundColor = Old;
			});

			Application.PrecacheModelRequested += new QuakeC.PrecacheFileRequestedEventHandler(delegate(object sender, QuakeC.PrecacheFileRequestedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.Green;
				Console.WriteLine("Precache model '{0}'", e);
				Console.ForegroundColor = Old;
			});

			Application.CVarChanged += new QuakeC.CVarChangedEventHandler(delegate(object sender, QuakeC.CVarChangedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.DarkYellow;
				Console.WriteLine("Change CVAR '{0}'", e);
				Console.ForegroundColor = Old;
			});

			Application.LightStyleChanged += new QuakeC.LightStyleChangedEventHandler(delegate(object sender, QuakeC.LightStyleChangedEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.Magenta;
				Console.WriteLine("Change light #{0} -> '{1}'", e.Light, e.Value);
				Console.ForegroundColor = Old;
			});

			Application.SpawnAmbientSound += new QuakeC.SpawnAmbientSoundEventHandler(delegate(object sender, QuakeC.SpawnAmbientSoundEventArgs e) {
				ConsoleColor Old = Console.ForegroundColor;
				Console.ForegroundColor = ConsoleColor.DarkRed;
				Console.WriteLine("Spawn ambient sound {0}", e);
				Console.ForegroundColor = Old;
			});


			// Load entities:
			Application.LoadEntities(Level, QuakeC.SkillLevel.Easy);*/

			//EdictCollection Edicts = new EdictCollection(Application);
			//Edicts.LoadFromLevel(Level);
			//Console.WriteLine(Edicts[0].Properties["classname"]);
			//Console.WriteLine(Edicts[0].Properties["classname"]);

			VirtualMachine VM = new VirtualMachine(Resources, Application);

			VM.DebugStringPrinted += new VirtualMachine.DebugStringPrintedEventHandler(delegate(object sender, VirtualMachine.DebugStringPrintedEventArgs e) { Console.Write(e); });
			VM.LightStyleChanged += new VirtualMachine.LightStyleChangedEventHandler(delegate(object sender, VirtualMachine.LightStyleChangedEventArgs e) { Console.WriteLine("LightStyleChanged: " + e); });
			VM.PrecacheModelRequested += new VirtualMachine.PrecacheFileRequestedEventHandler(delegate(object sender, VirtualMachine.PrecacheFileRequestedEventArgs e) { Console.WriteLine("Precache Model: " + e); });

			//VM.Invoke("test");

			VM.Invoke("main");

			VM.QuakeCSource.Globals["skill"].Value = 1;
			VM.LoadLevel(Level);

			
			//Console.WriteLine(VM.Edicts["worldspawn"]);

			Console.ReadKey(true);
		}
	}
}
