using System;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.Generic;

using Quake = id.Quake;
using Xna = Microsoft.Xna.Framework;
using System.Runtime.InteropServices;
using System.IO;

namespace XnaQuake.Renderer {
	class SkyRenderer : IDrawable, ILoadable {

		#region Private Members

		private BspRenderer.VertexPositionTextureDiffuseLightMap[] Vertices;
		private short[] Triangles;
		private Texture2D[] Textures;
		private readonly Quake.Files.Level Level;

		#endregion

		#region Public Properties

		private string skyName;
		public string SkyName {
			get { return this.skyName; }
			set { this.skyName = value; }
		}

		#endregion

		#region Constructor

		public SkyRenderer(Quake.Files.Level level, string skyName) {
			this.skyName = skyName;
			this.Level = level;
		}

		#endregion

		#region ILoadable Members

		private Xna.GraphicsDeviceManager Graphics;




		#endregion

		#region IDrawable Members

		public void Draw(RenderController controller) {

			if (this.Textures == null) return;

			bool IsQuake1Style = this.Textures.Length == 2;

			controller.Graphics.GraphicsDevice.RenderState.DepthBufferWriteEnable = false;

			if (IsQuake1Style) {
				controller.WorldEffect.CurrentTechnique = controller.WorldEffect.Techniques["QuakeSky"];
			} else {
				controller.WorldEffect.CurrentTechnique = controller.WorldEffect.Techniques["SkyBox"];
			}

			controller.WorldEffect.Begin();

			foreach (EffectPass Pass in controller.WorldEffect.CurrentTechnique.Passes) {
				Pass.Begin();

				if (IsQuake1Style) {
					controller.WorldEffect.Parameters["SkyTexture"].SetValue(this.Textures[0]);
					controller.WorldEffect.Parameters["CloudTexture"].SetValue(this.Textures[1]);
					controller.WorldEffect.CommitChanges();
				}

				for (int i = 0; i < this.Vertices.Length; i += 4) {

					if (!IsQuake1Style) {
						controller.WorldEffect.Parameters["SkyTexture"].SetValue(this.Textures[i >> 2]);
						controller.WorldEffect.CommitChanges();
					}

					controller.Graphics.GraphicsDevice.DrawUserIndexedPrimitives<BspRenderer.VertexPositionTextureDiffuseLightMap>(
						PrimitiveType.TriangleList,
						this.Vertices,
						i, 4, this.Triangles, 0, 2
					);
				}

				Pass.End();
			}

			controller.WorldEffect.End();


			controller.Graphics.GraphicsDevice.RenderState.DepthBufferWriteEnable = true;
		}

		#endregion

		#region ILoadable Members

		public void Load(id.Quake.ResourceLoader resources, Microsoft.Xna.Framework.GraphicsDeviceManager graphics, Xna.Content.ContentManager content) {
			this.Graphics = graphics;
			this.Triangles = new short[6] { 0, 1, 2, 2, 3, 0 };
				
			Xna.Vector3[] Positions = new Xna.Vector3[] {
				new Xna.Vector3(+1f,+1f,-1f), new Xna.Vector3(+1f,+1f,+1f), new Xna.Vector3(+1f,-1f,+1f), new Xna.Vector3(+1f,-1f,-1f),
				new Xna.Vector3(-1f,+1f,+1f), new Xna.Vector3(-1f,+1f,-1f),	new Xna.Vector3(-1f,-1f,-1f), new Xna.Vector3(-1f,-1f,+1f),
				new Xna.Vector3(+1f,+1f,+1f), new Xna.Vector3(-1f,+1f,+1f), new Xna.Vector3(-1f,-1f,+1f), new Xna.Vector3(+1f,-1f,+1f), 
				new Xna.Vector3(-1f,+1f,-1f), new Xna.Vector3(+1f,+1f,-1f), new Xna.Vector3(+1f,-1f,-1f), new Xna.Vector3(-1f,-1f,-1f),
				new Xna.Vector3(-1f,+1f,+1f), new Xna.Vector3(+1f,+1f,+1f), new Xna.Vector3(+1f,+1f,-1f), new Xna.Vector3(-1f,+1f,-1f),
				new Xna.Vector3(-1f,-1f,-1f), new Xna.Vector3(+1f,-1f,-1f), new Xna.Vector3(+1f,-1f,+1f), new Xna.Vector3(-1f,-1f,+1f),
			};
			Xna.Vector2[] TextureCoordinates = new Microsoft.Xna.Framework.Vector2[]{
				new Xna.Vector2(0f, 0f), new Xna.Vector2(1f, 0f), new Xna.Vector2(1f, 1f), new Xna.Vector2(0f, 1f)
			};
			this.Vertices = new BspRenderer.VertexPositionTextureDiffuseLightMap[Positions.Length];
			for (int i = 0; i < Positions.Length; ++i) {
				this.Vertices[i] = new BspRenderer.VertexPositionTextureDiffuseLightMap(Positions[i] * 10, TextureCoordinates[i & 3], Xna.Vector2.Zero);
			}

			switch (this.Level.FileVersion.GameVersion) {
				case id.Quake.Files.Level.Version.Game.Quake:
					foreach (Quake.Files.Level.TextureInfo Texture in this.Level.TextureInformation) {
						if (Texture.Name.StartsWith("sky")) {
							Quake.Files.Level.MipTexture SourceTexture = this.Level.MipTextures[Texture.TextureId];
							Quake.Files.CroppedMipMappedPicture CloudPlane = new Quake.Files.CroppedMipMappedPicture(SourceTexture, 0, 0, SourceTexture.Width / 2, SourceTexture.Height);
							Quake.Files.CroppedMipMappedPicture SkyPlane = new Quake.Files.CroppedMipMappedPicture(SourceTexture, SourceTexture.Width / 2, 0, SourceTexture.Width / 2, SourceTexture.Height);
							this.Textures = new Texture2D[]{
								Utility.TextureFromMipMappedPicture(graphics.GraphicsDevice, SkyPlane),
								Utility.TextureFromMipMappedPicture(graphics.GraphicsDevice, CloudPlane)
							};
							break;
						}
					}
					break;
				case id.Quake.Files.Level.Version.Game.Quake2:
					this.Textures = new Texture2D[6];
					string[] TextureNames = new string[] { "ft", "bk", "lf", "rt", "up", "dn" };
					for (int i = 0; i < Textures.Length; ++i) {
						string Filename = "env/" + this.SkyName + TextureNames[i] + ".pcx";
						this.Textures[i] = Utility.TextureFromPicture(graphics.GraphicsDevice, resources.LoadObject<Quake.Files.ZSoftPcx>(Filename).Picture);
					}
					break;
				default:
					throw new InvalidOperationException();
			}
		}

		public void Unload(Microsoft.Xna.Framework.GraphicsDeviceManager graphics, Microsoft.Xna.Framework.Content.ContentManager content) {
			if (this.Textures != null) {
				for (int i = 0; i < this.Textures.Length; ++i) {
					if (this.Textures[i] != null && !this.Textures[i].IsDisposed) this.Textures[i].Dispose();
				}
			}
		}

		#endregion
	}
}
